#ridge.coeffs() is the only function meant to be used directly by the user. 
#It calls all of the other functions in this script. 


##########################################
#Calcuate ridge regression estimates for a set of tuning parameters. 
#Returns a (p x length(lambda)) matrix of coefficient estimates. 
#The ith column contains the ridge estimates for the ith element of lambda. 
#If any of the elements in lambda are numerically zero and p>n-1, then they are ignored, 
#and the correponding results will be numerically very large. 
#lambda = scalar or vector of non-negative tuning parameters. 
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
#evec.crossprod.sx = pxp matrix of eigenvectors of the crossproduct of sx
#crossprod.sx.y = px1 matrix, the crossproduct of sx with y, which is the nx1 matrix of outcomes
#tol = positive numbers less than tol are numerically zero
##########################################

sbetas.ridge = function(lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,tol=sqrt(.Machine$double.eps)) {
	p = length(eval.crossprod.sx);
	results = matrix(sqrt(.Machine$double.xmax),p,length(lambda),dimnames=list(NULL,lambda));	
	if(max(min(lambda),min(eval.crossprod.sx))<tol) { 
		warning("Ignoring zero-valued lambdas for p > n-1");	
		pos.lambda = lambda>tol;
		lambda = lambda[pos.lambda];
	} else {
		pos.lambda = rep(T,length(lambda));
	}
	k = length(lambda);
	div = eval.crossprod.sx + rep(lambda, rep(p, k));
	a = drop(crossprod(evec.crossprod.sx,crossprod.sx.y))/div;
	dim(a) = c(p,k);
	results[,pos.lambda] = evec.crossprod.sx%*%a;	
	results;
}

##########################################
#Calculates the loglikelihood
#sbetas = length-p vector of values of the regression coefficients corresponding to the standardized design matrix, sx, and centered outcomes, y
#sigmasq = scalar, the error variance
#centered.y = length-n vector of centered outcomes
#sx = nxp design matrix, standardized (mean 0, unit variance)
##########################################
log.likelihood = function(sbetas,sigmasq,centered.y,sx) {
	-.5*sum((centered.y-sx%*%sbetas)^2)/sigmasq - .5*length(centered.y)*log(sigmasq)
}

##########################################
#Returns the sum of the parts of the h-loglihood (Lee and Nelder, 1996) that contain the tuning parameter lambda
#lambda = scalar or vector of non-negative tuning parameters. 
#sbetas = length-p vector of values of the regression coefficients corresponding to the standardized design matrix, sx, and centered outcomes, y
#sigmasq = scalar, the error variance
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
##########################################
maphl.lambda = function(lambda,sbetas,sigmasq,eval.crossprod.sx) {
	p = length(eval.crossprod.sx);
	k = length(lambda);
	lambda*sum(sbetas^2)/sigmasq - p*log(lambda) + colSums(log(matrix(rep(lambda,each=p)+rep(eval.crossprod.sx,times=k),nrow=p,ncol=k)))
}


##########################################
#chooses the lambda from a grid that minimizes cross-validation error.
#y = length-n vector of UNCENTERED ouctomes
#x = nxp design matrix, UNSTANDARDIZED
#n.folds = scalar, a positive integer for how many folds to do
#log.lambda.grid = list containing exactly three components: from, to, and length.out. 
#	From this, a grid of lambdas to evaluate is calculated through 'exp(do.call(seq,log.lambda.grid))'
##########################################
Kfold.cv = function(y,x,n.folds=5,log.lambda.grid) {
	
	p = ncol(x);
	folds = rep(1:n.folds,length=(n<-length(y)));
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	rss = 0;
	list.curr.new.sx = 
		list.centered.curr.new.y = 
		list.crossprod.curr.sx.curr.y = 
		list.eval.crossprod.curr.sx = 
		list.evec.crossprod.curr.sx = vector("list",n.folds);

	for(i in 1:n.folds) {
		curr.index = folds!=i
		curr.n = sum(curr.index);
		curr.x = x[curr.index,];
		curr.location.factor = colMeans(curr.x);
		curr.scale.factor = (colMeans((curr.x - rep(curr.location.factor,rep(curr.n, p)))^2))^.5;
		curr.sx = scale(curr.x,curr.location.factor,curr.scale.factor);
		curr.y = y[curr.index];
	
		pos.evals = 1:min(curr.n-1,p)
		beta0.hat = mean(curr.y);
		centered.curr.y = y[curr.index] - beta0.hat;

		list.curr.new.sx[[i]] = scale(matrix(x[!curr.index,],ncol=p),curr.location.factor,curr.scale.factor);
		list.centered.curr.new.y[[i]] = y[!curr.index] - beta0.hat;

		list.crossprod.curr.sx.curr.y[[i]] = crossprod(curr.sx,centered.curr.y);
		list.eval.crossprod.curr.sx[[i]] = ((eigen.decomp<-eigen(crossprod(curr.sx)))$values);
		list.eval.crossprod.curr.sx[[i]][-pos.evals] = 0;
		list.evec.crossprod.curr.sx[[i]] = eigen.decomp$vectors;
		rm(eigen.decomp);

		foo = sbetas.ridge(lambda.seq,list.eval.crossprod.curr.sx[[i]],list.evec.crossprod.curr.sx[[i]],list.crossprod.curr.sx.curr.y[[i]]);
		rss = rss + colSums((list.centered.curr.new.y[[i]] - list.curr.new.sx[[i]]%*%foo)^2);
	}
	curr.lambda = lambda.seq[which.min(rss)]
	log.lambda.grid$from = log(lambda.seq[max(1,which(lambda.seq==curr.lambda)-1)]);
	log.lambda.grid$to = log(lambda.seq[min(length(lambda.seq),which(lambda.seq==curr.lambda)+1)]);
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	rss = 0;
	for(i in 1:n.folds) {
		foo = sbetas.ridge(lambda.seq,list.eval.crossprod.curr.sx[[i]],list.evec.crossprod.curr.sx[[i]],list.crossprod.curr.sx.curr.y[[i]]);
		rss = rss + colSums((list.centered.curr.new.y[[i]] - list.curr.new.sx[[i]]%*%foo)^2);
	}
	lambda.seq[which.min(rss)]
}

##########################################
#Returns the MPML (Wecker and Ansley, 1983) or GMPML (Harville, 1977; Wahba, 1985)
#lambda = scalar or vector of non-negative tuning parameters. 
#centered.y = length-n vector of centered outcomes
#sx = nxp design matrix, standardized (mean 0, unit variance)
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
#evec.crossprod.sx = pxp matrix of eigenvectors of the crossproduct of sx
#crossprod.sx.y = px1 matrix, the crossproduct of sx with y, which is the nx1 matrix of outcomes
#adj = multiplicative scalar to adjust the penalty (differs between mpml and gmpml)
#tol = positive numbers less than tol are numerically zero;
##########################################
mpml.gmpml.model.fit = function(lambda,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,adj,tol=sqrt(.Machine$double.eps)) {
	k = length(lambda);
	p = length(eval.crossprod.sx);
	fitted.vals = sx%*%sbetas.ridge(lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y)
	loglik = log(colMeans(centered.y*(centered.y-fitted.vals)));
	penalty = - p*log(lambda)/adj + colSums(log(matrix(rep(lambda,each=p)+rep(eval.crossprod.sx,times=k),nrow=p,ncol=k)))/adj;
	loglik + penalty
}

##########################################
#chooses the lambda from a grid that maximizes the MPML or the GMPML
#mpml = logical. If T, maximizes the mpml. If F, maximizes the GMPML
#centered.y = length-n vector of centered outcomes
#sx = nxp design matrix, standardized (mean 0, unit variance)
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
#evec.crossprod.sx = pxp matrix of eigenvectors of the crossproduct of sx
#log.lambda.grid = list containing exactly three components: from, to, and length.out. 
#	From this, a grid of lambdas to evaluate is calculated through 'exp(do.call(seq,log.lambda.grid))'
##########################################
find.opt.mpml.gmpml = function(mpml=T,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid) {
	n = length(centered.y);
	p = length(eval.crossprod.sx);
	adj = ifelse(mpml,n,n-1);
	pos.eval.crossprod.sx = eval.crossprod.sx[eval.crossprod.sx>0];

	lambda.seq = exp(do.call(seq,log.lambda.grid));
	criteria = mpml.gmpml.model.fit(lambda.seq,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,adj)
	curr.lambda = lambda.seq[which.min(criteria)]
	log.lambda.grid$from = log(lambda.seq[max(1,which(lambda.seq==curr.lambda)-1)]);
	log.lambda.grid$to = log(lambda.seq[min(length(lambda.seq),which(lambda.seq==curr.lambda)+1)]);
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	criteria = mpml.gmpml.model.fit(lambda.seq,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,adj)
	curr.lambda = lambda.seq[which.min(criteria)]
	
	return.ests = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y)
	list(coef=return.ests,lambda=curr.lambda)
}



##########################################
###Penalized Model Fit Methods#####
###MPML/GMPML methods
#Returns a 'model fit + penalty' value for various penalties. 
#lambda = scalar or vector of non-negative tuning parameters. 
#penalty.func = character. Can be 'lr', 'aicC', 'bic', 'gcv', 'gcvC', or 'rgcv'. 
#centered.y = length-n vector of centered outcomes
#sx = nxp design matrix, standardized (mean 0, unit variance)
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
#evec.crossprod.sx = pxp matrix of eigenvectors of the crossproduct of sx
#crossprod.sx.y = px1 matrix, the crossproduct of sx with y, which is the nx1 matrix of outcomes
#n = number of observations. 
#piecewise = logical. Return the individual 'model fit' and 'penalty' terms or just there sum?
#tol = positive numbers less than tol are numerically zero
##########################################
pen.model.fit = function(lambda,penalty.func,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,n,piecewise=F,tol=sqrt(.Machine$double.eps)) {
	fitted.vals = sx%*%sbetas.ridge(lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y)
	k = length(lambda);		
	p = length(eval.crossprod.sx);
	if(penalty.func=="rgcv") {
		rep.lambda = rep(lambda,each=p)
		rep.eval.crossprod.sx = rep(eval.crossprod.sx,times=k)	
		penalty.arglist = list(df=1+colSums(matrix(1/(1+rep.lambda/rep.eval.crossprod.sx),nrow=p,ncol=k)),df2=1+colSums(matrix(1/(1+(rep.lambda/rep.eval.crossprod.sx)^2),nrow=p,ncol=k)),n=n);
		rm(rep.lambda,rep.eval.crossprod.sx);
	} else if(penalty.func=="lr") {
		penalty.arglist = list(df=-p*log(lambda)+colSums(log(matrix(rep(lambda,each=p)+rep(eval.crossprod.sx,times=k),nrow=p,ncol=k))),n=n);
	} else {	
		penalty.arglist = list(df=1+colSums(matrix(1/(1+rep(lambda,each=p)/rep(eval.crossprod.sx,times=k)),nrow=p,ncol=k)),n=n);
	}
	loglik = log(colMeans((centered.y-fitted.vals)^2));
	penalty = do.call(penalty.func,penalty.arglist);
	if(piecewise) {c(loglik=loglik,penalty=penalty)} else {loglik+penalty}
}

##########################################
#chooses the lambda from a grid that maximizes 'model fit+penalty'
#penalty.func = character. Can be 'lr', 'aicC', 'bic', 'gcv', 'gcvC', or 'rgcv'. 
#centered.y = length-n vector of centered outcomes
#sx = nxp design matrix, standardized (mean 0, unit variance)
#eval.crossprod.sx = length-p vector of eigenvalues of the crossproduct of sx, where sx is a
#	standardized (mean 0, unit variance), nxp design matrix. 
#evec.crossprod.sx = pxp matrix of eigenvectors of the crossproduct of sx
#log.lambda.grid = list containing exactly three components: from, to, and length.out. 
#	From this, a grid of lambdas to evaluate is calculated through 'exp(do.call(seq,log.lambda.grid))'
##########################################
find.opt.model = function(penalty.func,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid) {
	n = length(centered.y);
	p = length(eval.crossprod.sx);
	pos.eval.crossprod.sx  = eval.crossprod.sx[eval.crossprod.sx > 0];

	lambda.seq = exp(do.call(seq,log.lambda.grid));
	criteria = pen.model.fit(lambda.seq,penalty.func,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,n)
	curr.lambda = lambda.seq[which.min(criteria)]
	log.lambda.grid$from = log(lambda.seq[max(1,which(lambda.seq==curr.lambda)-1)]);
	log.lambda.grid$to = log(lambda.seq[min(length(lambda.seq),which(lambda.seq==curr.lambda)+1)]);
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	criteria = pen.model.fit(lambda.seq,penalty.func,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,n)
	curr.lambda = lambda.seq[which.min(criteria)]

	min.criteria = pen.model.fit(curr.lambda,penalty.func,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,n,piecewise=T)
	return.ests = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);
	list(coef=return.ests,lambda=curr.lambda,opt.fit = min.criteria[1],opt.penalty = min.criteria[2])
}

#Following are various functions to determine a penalty

lr = function(df,n) {
	2*df/n
}

aicC = function(df,n) {
	2*(df+1)/pmax(.Machine$double.eps^.5,n-df-2);
}

bic = function(df,n) {
	log(n)*(df+1)/n;
}

gcv = function(df,n) {
	-2*log(1-df/n);
}

gcvC = function(df,n) {
	df = pmin(n-.Machine$double.eps^.5,df + 1);
	-2*log(1-df/n);
}

rgcv = function(df,df2,n,gamma=0.3) {
	-2*log(1-df/n)+log(gamma+(1-gamma)*df2/n)
}


##########################################
#Primary function. Input the training data, validation data, and an intial
#grid of lambdas to test (which will be further
#explored at a higher resolution) to return
#the ridge regression estimates and predicition error estimates
#x = nxp design matrix, UNSTANDARDIZED
#sx = nxp design matrix, STANDARDIZED (mean 0, unit variance)
#y = length-n vector of UNCENTERED ouctomes
#new.sx = A new design matrix (independent of sx) that is scaled 
#	exactly the same way as sx
#new.y = a new vector of outcomes
#location.factor = length-p vector which was used to center x
#scale.factor = length-p vector which was used to scale x
#	ie sx = scale(x,location.factor,scale.factor);
#log.lambda.grid = list containing exactly three components: from, to, and length.out. 
#	From this, a grid of lambdas to evaluate is calculated through 'exp(do.call(seq,log.lambda.grid))'
#	The grid search process is iterative, meaning each method will home in on the choice of lambda
#R2.guess = scalar in (0,1). What do you guess that R2 is? Used by Hyp1. 
#R2 = scalar in (0,1). What is the true R2? Of course you don't know this for real data, so you can
#	just plug in anything here. Used by Hyp2 to determine its biased guess of R2. 
##########################################
ridge.coeffs = function(x,sx,y,new.sx,new.y,
			location.factor,scale.factor,
			log.lambda.grid,
			R2.guess,R2) 
	{

	meth = c("opt","5-cv","gcv","bic","aicC","rgcv","gcvC","mpml","gmpml","maphl","lr","hyp1","hyp2");

	n = nrow(sx); p = ncol(sx);
	pos.evals = 1:min(n-1,p)
	beta0.hat = mean(y);
	centered.y = drop(y) - beta0.hat;
	centered.new.y = new.y - beta0.hat;

	crossprod.sx.y = crossprod(sx,centered.y);
	eval.crossprod.sx = ((eigen.decomp<-eigen(crossprod(sx)))$values);
	eval.crossprod.sx[-pos.evals] = 0;
	pos.eval.crossprod.sx = eval.crossprod.sx[pos.evals];
	evec.crossprod.sx = eigen.decomp$vectors;
	rm(eigen.decomp);

	
	coeffs = matrix(0,nrow=p,ncol=length(meth));
	lambda.mspe = matrix(0,nrow=length(meth),ncol=2);
	colnames(coeffs) = rownames(lambda.mspe) = meth;
	#mspe is the average squared prediction error between ynew and xnew%*%betahat
	colnames(lambda.mspe) = c("lambda","mspe");

	gamma.params = matrix(0,nrow=2,ncol=4,dimnames=list(c("hyp1","hyp2"),c("shape","rate","mean","var")))
 
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	if(min(eval.crossprod.sx)<sqrt(.Machine$double.eps)) { 
		log.lambda.grid$from = log(min(lambda.seq[lambda.seq>sqrt(.Machine$double.eps)]));
		lambda.seq = exp(do.call(seq,log.lambda.grid));
	}
	new.log.lambda.grid = log.lambda.grid;

	#####
	##TRUE OPTIMAL LAMBDA
	curr.lambda = lambda.seq[which.min(colMeans((centered.new.y-new.sx%*%sbetas.ridge(lambda.seq,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y))^2))]
	new.log.lambda.grid$from = log(lambda.seq[max(1,which(lambda.seq==curr.lambda)-1)]);
	new.log.lambda.grid$to = log(lambda.seq[min(length(lambda.seq),which(lambda.seq==curr.lambda)+1)]);
	lambda.seq = exp(do.call(seq,new.log.lambda.grid));
	curr.lambda = lambda.seq[which.min(colMeans((centered.new.y-new.sx%*%sbetas.ridge(lambda.seq,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y))^2))]
	 
	coeffs[,"opt"] = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);
	lambda.mspe["opt","lambda"] = curr.lambda;

	#####
	##5-fold CV
	lambda.mspe["5-cv","lambda"] = Kfold.cv(y,x,5,log.lambda.grid)
	coeffs[,"5-cv"] = sbetas.ridge(lambda.mspe["5-cv","lambda"],eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);

	#####
	##GCV
	gcv.results = find.opt.model("gcv",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"gcv"] = gcv.results$coef;
	lambda.mspe["gcv","lambda"] = gcv.results$lambda;

	#####
	##GCV (counting sigma^2 in the degrees of freedom);
	gcv.results = find.opt.model("gcvC",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"gcvC"] = gcv.results$coef;
	lambda.mspe["gcvC","lambda"] = gcv.results$lambda;

	#####
	##rGCV
	rgcv.results = find.opt.model("rgcv",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"rgcv"] = rgcv.results$coef;
	lambda.mspe["rgcv","lambda"] = rgcv.results$lambda;

	#####
	#AICc
	aicC.results = find.opt.model("aicC",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"aicC"] = aicC.results$coef;
	lambda.mspe["aicC","lambda"] = aicC.results$lambda;

	#####
	#BIC
	bic.results = find.opt.model("bic",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"bic"] = bic.results$coef;
	lambda.mspe["bic","lambda"] = bic.results$lambda;

	#######
	#MPML
	mpml.results = find.opt.mpml.gmpml(mpml=T,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"mpml"] = mpml.results$coef;
	lambda.mspe["mpml","lambda"] = mpml.results$lambda;

	#######
	#GMPML
	gmpml.results = find.opt.mpml.gmpml(mpml=F,centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid)
	coeffs[,"gmpml"] = gmpml.results$coef;
	lambda.mspe["gmpml","lambda"] = gmpml.results$lambda;

	#####
	##MAPHL
	lambda.seq = exp(do.call(seq,log.lambda.grid));
	curr.sigmasq = 1;
	curr.lambda = 1;
	j=0;old.lambda = 0;

	while(abs(log(curr.lambda/old.lambda))>1e-3|j<=25) {
		j=j+1;
		curr.sbetas = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);		
		curr.sigmasq = (sum((centered.y-sx%*%curr.sbetas)^2)+curr.lambda*sum(curr.sbetas^2))/(n-1);
		old.lambda = curr.lambda;
		curr.lambda = lambda.seq[which.min(maphl.lambda(lambda.seq,curr.sbetas,curr.sigmasq,eval.crossprod.sx))];		
		new.log.lambda.grid$from = max(log.lambda.grid$from,log(lambda.seq[which(lambda.seq==curr.lambda)])-(j<=20)-j/5);
		new.log.lambda.grid$to = min(log.lambda.grid$to,log(lambda.seq[which(lambda.seq==curr.lambda)])+(j<=20)+j/5);
		lambda.seq = exp(do.call(seq,new.log.lambda.grid));
	}

	coeffs[,"maphl"] = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);
	lambda.mspe["maphl","lambda"] = curr.lambda;

	#####
	#LR
	lr.results = find.opt.model("lr",centered.y,sx,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y,log.lambda.grid);
	coeffs[,"lr"] = lr.results$coef;
	lambda.mspe["lr","lambda"] = lr.results$lambda;

	###HYP1 (with guess of R2 from a distribution)
	#lambda ~ G(shape.gam,rate.gam)

	shape.gam = p/2;##Mean(Lambda) = shape/rate;	
	rate.gam = R2.guess/(1-R2.guess)/2;##Var(Lambda) = shape/rate^2;	
	gamma.params["hyp1",] = c(shape.gam,rate.gam,shape.gam/rate.gam,shape.gam/rate.gam^2);

	curr.sbetas = coeffs[,"gmpml"];
	curr.lambda = 1;log.curr.lambda=0;
	j=0;old.lambda=0;
	
	while(abs(log(curr.lambda/old.lambda))>1e-3) {
		j=j+1;
		old.lambda = curr.lambda;
		curr.sigmasq = (sum((centered.y-sx%*%curr.sbetas)^2)+curr.lambda*sum(curr.sbetas^2))/(n+p+2);	
		curr.lambda = (p+2*shape.gam-2)/(sum(curr.sbetas^2)/curr.sigmasq+2*rate.gam);
		log.curr.lambda = log(curr.lambda);
		if(log.curr.lambda>=log.lambda.grid$to) {curr.lambda = old.lambda = exp(log.lambda.grid$to);}
		if(log.curr.lambda<=log.lambda.grid$from) {curr.lambda = old.lambda = exp(log.lambda.grid$from);}
		curr.sbetas = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);	
	}

	coeffs[,"hyp1"] = curr.sbetas;
	lambda.mspe["hyp1","lambda"] = curr.lambda;
	rm(shape.gam,rate.gam);

	###HYP2 (with upwardly biased estimate of R2)
	#lambda ~ G(shape.gam,rate.gam)

	shape.gam = p/2;##Mean(Lambda) = shape/rate;	
	rate.gam = (R2+0.1)/(1-(R2+0.1))/2;##Var(Lambda) = shape/rate^2;	
	gamma.params["hyp2",] = c(shape.gam,rate.gam,shape.gam/rate.gam,shape.gam/rate.gam^2);

	curr.sbetas = coeffs[,"gmpml"];
	curr.lambda = 1;log.curr.lambda=0;
	j=0;old.lambda=0;
	
	while(abs(log(curr.lambda/old.lambda))>1e-3) {
		j=j+1;
		old.lambda = curr.lambda;
		curr.sigmasq = (sum((centered.y-sx%*%curr.sbetas)^2)+curr.lambda*sum(curr.sbetas^2))/(n+p+2);	
		curr.lambda = (p+2*shape.gam-2)/(sum(curr.sbetas^2)/curr.sigmasq+2*rate.gam);
		log.curr.lambda = log(curr.lambda);
		if(log.curr.lambda>=log.lambda.grid$to) {curr.lambda = old.lambda = exp(log.lambda.grid$to);}
		if(log.curr.lambda<=log.lambda.grid$from) {curr.lambda = old.lambda = exp(log.lambda.grid$from);}
		curr.sbetas = sbetas.ridge(curr.lambda,eval.crossprod.sx,evec.crossprod.sx,crossprod.sx.y);	
	}

	coeffs[,"hyp2"] = curr.sbetas;
	lambda.mspe["hyp2","lambda"] = curr.lambda;
	rm(shape.gam,rate.gam);


	lambda.mspe[,"mspe"] = colMeans((centered.new.y-new.sx%*%coeffs)^2)

	list(coeffs = coeffs,lambda.mspe=lambda.mspe, gamma.params=gamma.params)

}


