#This contains all functions required to run a simulation. 
#The function run.ch4.sim.wrapper is meant to be the primary function here. 


##############################
#Random draw of betas as described in the manuscript
#p = scalar, the length of the desired beta
#AR.pi = scalar in (0,1), the larger this is, the more correlated neighboring elements of beta will be. 
##############################
draw.betas = function(p,AR.pi) {
	betas = numeric(p);
	#The quantity k = #{Z:Z!=3} comes from a conditional binomial distribution. ie
	#the number of elements of beta that don't come from the normal mixture component
	#follows a Binomial(p,0.01) distribution, conditional on at least 3 "successes";
	if(p<=3) {
		k = 3;
	} else if (p<=100) {
		k = sample(3:p,1,prob=dbinom(3:p,p,.01)/pbinom(2,p,.01,lower.tail=F))
	} else {
		#If p is large, then, to save time, ignore the probability of big values of k, which will be tiny.
		k = sample(3:round(p/5),1,prob=dbinom(3:round(p/5),p,.01)/pbinom(2,p,.01,lower.tail=F))	
	}
	#Given k, choose the (latent) component membership
	z = c(sample(1:2,k,r=T),rep(3,p-k))[sample(p)]
	betas[which(z==3)] = rnorm(sum(z==3),sd=.001);
	betas[which(z==2)] = rexp(sum(z==2));
	betas[which(z==1)] = rt(sum(z==1),df=3)/sqrt(3);
	drop(betas%*%toeplitz(AR.pi^(0:(p-1))));
}

##############################
#Solve for sigmasq. Returns a scalar
#betas = vector of length p
#chol.varcov.X = upper triangular pxp matrix, the cholesky decomposition of the variance-covariance matrix of the covariates
#R2 = scalar in (0,1), the true coefficient of determination
##############################
calc.sigmasq = function(betas,chol.varcov.X,R2) {
	(1/R2 - 1)*sum((chol.varcov.X%*%betas)^2)
}

##############################
###Draw x,y. Returns a list relevant quantities to do a regression on. 
#n = positive integer, sample size of training data
#beta0 = scalar, the regression model intercept
#betas = vector of length p, the regression coefficients
#reg.errors = list containing 'dist', which is a character vector giving the name of an R function to sample a distribution,
#	'args', which is a list of arguments to pass to 'dist', and 'offset', which is a constant amount by which to offset
#	the resulting draws. This specifies the form of the error term in the outcome model. 
#mu.X = vector of length p, the mean of the covariates
#chol.varcov.X = upper triangular pxp matrix, the cholesky decomposition of the variance-covariance matrix of the covariates
#X.errors = list, similar to 'reg.errors', but for the covariates. The covariates are generated
#	by calculating the X.errors, scaling by chol.varcov.X, and adding mu.X. 
#mixture.X = matrix with p columns. If nonnull, each observation in x is shifted by a randomly selected column from this matrix
#n.new = positive integer, sample size of validation data
#location.factor = length-p vector to be used to shift x. If null, it will be calculated as the column means of the training data
#scale.factor = length-p vector to be used to scale x. If null, it will be caluclated as the marginal std deviations of the training data
##############################
draw.xy = function(n,beta0,betas,sigmasq=1,reg.errors=list(dist="rnorm",args=list(mean=0,sd=1),offset=0),
				mu.X=numeric(length(betas)),chol.varcov.X=diag(1,length(betas)),
				X.errors=list(dist="rnorm",args=list(mean=0,sd=1)),mixture.X=NULL,
				n.new=10,location.factor=NULL,scale.factor=NULL) {

		p = length(betas);
	
		x = rep(mu.X,each=n) + matrix(do.call(X.errors$dist,c(n=n*p,X.errors$args)),ncol=p)%*%chol.varcov.X;
		new.x = rep(mu.X,each=n.new) + matrix(do.call(X.errors$dist,c(n=n.new*p,X.errors$args)),ncol=p)%*%chol.varcov.X;
		
		if(!is.null(mixture.X)) {
			if(class(mixture.X)!="matrix"|ncol(mixture.X)!=p) {stop("mixture.X must be a matrix with p columns");}
			x = x + mixture.X[sample(nrow(mixture.X),n,T),]
			new.x = new.x + mixture.X[sample(nrow(mixture.X),n.new,T),]
		}

		#Outcomes
		y =  beta0 + x%*%betas + sqrt(sigmasq)*(do.call(reg.errors$dist,c(n=n,reg.errors$args))+reg.errors$offset);
		new.y = beta0 + drop(new.x%*%betas) + sqrt(sigmasq)*(do.call(reg.errors$dist,c(n=n.new,reg.errors$args))+reg.errors$offset);

		if(p==1) {x = matrix(x);}
		
		if(is.null(location.factor)) {location.factor = colMeans(x);}
		if(is.null(scale.factor)) {scale.factor = (colMeans((x - rep(location.factor,rep(n, p)))^2))^.5;}

		sx = scale(x,center=location.factor,scale=scale.factor);
		new.sx = scale(new.x,center=location.factor,scale=scale.factor);

		list(x=x,y=y,
			new.x=new.x,new.y=new.y,
			sx=sx,new.sx=new.sx,
			location.factor=location.factor,scale.factor=scale.factor,
			errors=y-drop(beta0-x%*%betas),new.errors=new.y-beta0-drop(new.x%*%betas));
	
}

##############################
#Wrapper function. Takes a list of arguments, 'params', and sends it to the run.sim function. 
##############################
run.sim.wrapper <- function(params) {
	run.sim(nsim=params$nsim,n=params$n,n.new=params$n.new,
				beta0=params$beta0,betas=params$betas,R2=params$R2,
				rho=params$rho,sigmasq.X=params$sigmasq.X,corr.structure=params$corr.structure,
				sim.number=params$sim.number,beta.label=params$beta.label,
				verbose=params$verbose,
				random.seed=params$random.seed,
				betas.func=params$betas.func,
				perturb.corr.X=params$perturb.corr.X)
}
	
##############################
#Runs a simulation for a particular setting. Returns a list of components giving various results, including:
#--colmeans.scaled.mspe is a vector of the average rMSPE for each method.
#--scaled.mspe is a matrix of every single rMSPE from every simulation.
#--mspe is a matrix of the standard MSPE from every simulation.
#--lambda is a matrix of the selected lambda from every simulation
#--R2.guess is a vector of the automated guesses of R2 made by HYP1
#--snr is a vector of the signal to noise ratio from each simulation: beta^top*beta/sigmasq
#--gamma.means is a matrix of the induced means for lambda from hyp1 and hyp2. 
#Arguments:
#nsim = positive integer, number of simulations to run
#n = positive integer, sample size of training data
#n.new = positive integer, sample size of validation data
#beta0 = scalar, the regression model intercept
#betas = vector of length p, the regression coefficients
#R2 = scalar in (0,1), the true coefficient of determination
#rho = scalar in (0,1), depending on 'corr.struct', determines the correlation between covariates
#sigmasq.X = positive scalar, multiply the covariance matrix of X by this number
#corr.struct = character, one of 'AR1' (autoregressive), 'CS' (compound symmetrix), 'bCS' (blockwise compound symmetric). Otherwise, 
#	the correlation matrix is assumed diagonal. 
#sim.number = scalar, whatever number you want; this just gets passed to the output. For keeping track of various settings. 
#beta.label = character, whatever you want; this just gets passed to the output. For keeping track of various settings. 
#verbose = logical. Do you want to get back lots and lots of details? Can result in big files. 
#random.seed = positive integer. Seed for the random number generator.  
#betas.func = optional list to sample beta from a distribution rather than fix it. If supplied, must contain 'func', which is a 
#	function for sampling the betas, and any required arguments. 
#perturb.corr.X = optional list to perturb the correlation matrix, as described in Appendix 7.5. If supplied, must contain 
#	'M', which is a positive integer, 'mean', whichs is the mean of the perturbing normal distribution, and 'delta', 
#	which is a scalar in (0,1-rho). 
##############################
run.sim <- function(nsim,n,n.new=n.new,
				beta0,betas,R2,
				rho,sigmasq.X,corr.structure,
				sim.number,beta.label,
				verbose,
				random.seed=NULL,
				betas.func=NULL,
				perturb.corr.X=NULL) {


	if(!is.null(random.seed)) {set.seed(as.numeric(random.seed));}
	if(class(betas)=="list") {p=betas.func$p;
	} else {p=length(betas);}

	meth = c("opt","5-cv","gcv","bic","aicC","rgcv","gcvC","mpml","gmpml","maphl","lr","hyp1","hyp2");

	lambda = mspe = cmspe = matrix(0,nrow=nsim,ncol=length(meth));
	colnames(lambda) = colnames(mspe) = colnames(cmspe) = meth;
	if(verbose) {results = array(0,c(p,length(meth),nsim));colnames(results) = meth;}

	R2.guesses = numeric(nsim);
	snr = numeric(nsim);
	gamma.means = matrix(0,nrow=nsim,ncol=2);
	colnames(gamma.means) = c("ga","ga_wrongR2");

	if(corr.structure == "AR1") {#Autoregressive/Toeplitz
		corr.X = toeplitz(rho^(0:(p-1)));
	} else if(corr.structure == "CS") {#Compound Symmetric
		corr.X = (matrix(rho,nrow=p,ncol=p) + diag(1-rho,p))
	} else if(corr.structure =="bCS") {#Blockwise Compound Symmetric
		p.div = rep(floor(p/10),10);
		remainder.index = sample(10,p - sum(p.div));
		p.div[remainder.index] = p.div[remainder.index] + 1;
		corr.X = bdiag(mapply(matrix,nrow=p.div,ncol=p.div,MoreArgs=list(data=rho),SIMPLIFY=F))+diag(1-rho,p);
		rm(p.div,remainder.index);
	} else {
		corr.X = diag(1,p);
	}

	chol.varcov.X = sqrt(sigmasq.X)*chol(corr.X);
	
	if(class(betas)=="numeric"&is.null(perturb.corr.X)) {
		sigmasq = calc.sigmasq(betas,chol.varcov.X,R2);
	}

	begin = Sys.time();

	for(i in 1:nsim) {

		if(class(perturb.corr.X)=="list") {
			Ut = matrix(rnorm(p*perturb.corr.X$M,mean=perturb.corr.X$mean),nrow=p,ncol=perturb.corr.X$M);
			Ut = Ut/sqrt(rowSums(Ut^2));
			chol.varcov.X = sqrt(sigmasq.X)*chol(corr.X + perturb.corr.X$delta*(tcrossprod(Ut) - diag(1,p)));
			rm(Ut);
		} 

		if(class(betas.func)=="list") {
			betas = do.call(betas.func$func,list(betas.func$p,betas.func$AR.pi));
		}

		if(class(betas.func)=="list"|class(perturb.corr.X)=="list") {
			sigmasq = calc.sigmasq(betas,chol.varcov.X,R2);
		}

		dat = draw.xy(n,beta0,betas,sigmasq,
				chol.varcov.X=chol.varcov.X,
				n.new=n.new)
		x = dat$x;sx = dat$sx;y = dat$y;
		new.sx = dat$new.sx;new.y = dat$new.y;
		location.factor = dat$location.factor;
		scale.factor = dat$scale.factor;

		R2.guess = max(.01,R2 + runif(1,-0.05,0.15));
		
		log.lambda.min = log(p) + log(1/R2.guess-1) - 7;
		log.lambda.max = log.lambda.min + 14;	
		log.lambda.grid = list(from=log.lambda.min,to=log.lambda.max,length.out=50);

		foo = ridge.coeffs(x,sx,y,new.sx,new.y,location.factor,scale.factor,log.lambda.grid,R2.guess,R2)
		lambda[i,meth] = foo$lambda.mspe[meth,1];
		mspe[i,meth] = foo$lambda.mspe[meth,2];
		R2.guesses[i] = R2.guess;
		snr[i] = sum(betas^2)/sigmasq;
		gamma.means[i,] = foo$gamma.params[,"mean"]
		if(verbose) {results[,,i] = foo$coeffs/scale.factor;}

		if(i%%10==0) {cat(sim.number,"-",i," proj. run time:",round(proj.run.time<-nsim*difftime(Sys.time(),begin,units="hours")/i,2),"hours. proj. end time: ",format.Date(begin  + proj.run.time),"\n\n",file="tracktime.txt",append=T)}

	}
	Sys.time() - begin;

	if(!verbose) {results=NULL;}
	scaled.mspe = 1e2*(mspe[,-1]/mspe[,1]-1);
	colmeans.scaled.mspe = colMeans(scaled.mspe);
	list(regression.params = list(n=n,beta0=beta0,betas=betas.func,R2=R2,sigmasq=ifelse(class(betas.func)=="list",0,sigmasq)),
		x.params = list(rho=rho,sigmasq.X=sigmasq.X,corr.structure=corr.structure,perturb.corr.X=perturb.corr.X),
		simulation.params = list(nsim=nsim,beta.label=beta.label),
		colmeans.scaled.mspe = colmeans.scaled.mspe,
		lambda=lambda,
		mspe=mspe,
		scaled.mspe=scaled.mspe,
		results=results,
		R2.guess = R2.guesses,
		snr=snr, 
		gamma.means = gamma.means)

}






